import re, string
import socket, urllib
import os, signal, glob
import struct, math, operator
import ConfigParser
import types

# Use simplejson or Python 2.6 json, prefer simplejson.
try: import simplejson as json
except ImportError: import json

# Crython version (Of course, this can be different on the PC and CryoBoard
# sides. This variable is provided to answer questions about mismatches.)
CRYTHON_VERSION = 1

# Default ports
DEFAULT_CGI_PORT = 80
LOCAL_STREAMER_PORT = 9025

# Number of channels
NUM_RAW_CHANNELS = 64
NUM_COOKED_CHANNELS = 16
NUM_HEATER_CHANNELS = 8

# Sampling rate (IMPORTANT: please leave this as floating point! It's
# necessary to keep unit conversions from truncating.)
FSAMP = 100.0

# Channel parameters
SQUARE = 0
SINE = 1
DC = 2

# ADC PGA settings. Key is gain, value is register setting
PGA_GAIN_TO_REG_MAP = {
	1 : 0,
	2 : 1,
	4 : 2,
	8 : 3,
	16 : 4,
	32 : 5,
	64 : 6
}
PGA_REG_TO_GAIN_MAP = dict([(v,k) for (k,v) in PGA_GAIN_TO_REG_MAP.iteritems()])

class CryoError(Exception):
	pass

class CryoNetworkError(CryoError):
	pass

class CryoInternalError(CryoError):
	pass

class CryoArgumentError(CryoError):
	pass

class CryoStreamerError(CryoError):
	pass

class CryoRemoteError(CryoError):
	pass

class CryoCalibrationError(CryoError):
	pass


def require(arg_name, *allowed):
	'''
	This is a decorator based on code from here:
	
		http://code.activestate.com/recipes/454322/
	
	It takes care of type checking, which is important since these methods
	are available remotely.

	Decorators are neat, but if they scare you, feel free to ignore them.
	'''

	def make_wrapper(f):
		if hasattr(f, "wrapped_args"):
			wrapped_args = getattr(f, "wrapped_args")
		else:
			code = f.func_code
			wrapped_args = list(code.co_varnames[:code.co_argcount])

		try:
			arg_index = wrapped_args.index(arg_name)
		except ValueError:
			raise NameError, arg_name

		def wrapper(*args, **kwargs):
			if len(args) > arg_index:
				arg = args[arg_index]

				# Type comparison
				allowed_types = filter(lambda x: type(x) == types.TypeType, allowed)
				allowed_values = [ x for x in allowed if x not in allowed_types ]

				if not isinstance(arg, allowed_types) and arg not in allowed_values:

					type_list = " or ".join(str(allowed_type) for allowed_type in allowed_types)
					raise TypeError, "Expected '%s' to be %s; was %s(%s)." % (arg_name, type_list, type(arg), str(arg))

			else:
				if arg_name in kwargs:
					arg = kwargs[arg_name]

					allowed_types = filter(lambda x: type(x) == types.TypeType, allowed)
					allowed_values = [ x for x in allowed if x not in allowed_types ]

					if not isinstance(arg, allowed_types) and arg not in allowed_values:
						type_list = " or ".join(str(allowed_type) for allowed_type in allowed_types)
						raise TypeError, "Expected '%s' to be %s; was %s(%s)." % (arg_name, type_list, type(arg), str(arg))

			return f(*args, **kwargs)

		wrapper.wrapped_args = wrapped_args
		return wrapper

	return make_wrapper


class Units(object):
	# Machine units
	RAW = "Raw"

	HZ = "Hertz"

	VOLTS = "Volts"
	#AMPS = "Amps"
	#WATTS = "Watts"

	KELVIN = "Kelvin"

	DAC_COUNTS = "DAC Counts"
	ADC_COUNTS = "ADC Counts"

	OHMS = "Ohms"


class CryoBoard(object):
	'''
	Cryo board!

	This is the master class for using the Winterland Cryoelectronics
	board. You should instantiate one of these for each board you intend
	to interact with, regardless of whether those interactions are local
	or not. The class is supposed to be smart enough to figure it out.
	'''

	def __init__(self, hostname='localhost', port=DEFAULT_CGI_PORT):
		'''
		Construct a cryo board. You must supply a hostname for the
		board; if you're running Python code on the board itself, you
		may not a-priori know (or wish to code in) the IP address. In
		this case, it's sufficient to use 'localhost' or '127.0.0.1'.
		'''

		# Uncomment this for debugging purposes (to force network
		# access instead of local access)
		#self._is_local = False

		self._hostname = hostname
		if self.is_local():
			self._proxy = CryoLocalProxy()
		else:
			self._proxy = CryoRemoteProxy(hostname, port=port)
	

	def __getattribute__(self, name):
		'''
		This is how we delegate methods to either socket or hardware
		calls. In __init__, we figured out whether to use local calls
		or not, and instantiated a proxy object depending on which.
		Here, we delegate any CryoBoard::do_something() methods (or
		class variable accesses) to the proxy for handling. It's
		magic (except that you're reading this because it's not
		working properly. Sorry.)
		'''

		try:
			return object.__getattribute__(self, name)
		except AttributeError:
			proxy = object.__getattribute__(self, '_proxy')
			return proxy.__getattribute__(name)


	def get_hostname(self):
		'''
		Return the hostname used to generate this board
		'''
		return self._hostname


	def get_ip(self):
		'''
		Determines the IP address for this board
		'''

		# This is used to match IP addresses
		if not hasattr(self, '_ipre'):
			self._ipre = re.compile(r"""
				\b(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.
				(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.
				(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.
				(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\b
				""", re.VERBOSE)

		hostname = self.get_hostname()

		# If we were given an IP address, use it.
		if(self._ipre.match(hostname)):
			self._ip = hostname
			return self._ip

		# Otherwise, resolve the hostname
		try:
			ip = socket.gethostbyname(hostname)
		except socket.gaierror:
			raise CryoNetworkError("Failed to resolve hostname %s" % hostname)

		self._ip = ip
		return self._ip


	def is_local(self):
		'''
		Determines whether this CryoRemote object will talk to local
		hardware (i.e. we can use direct hardware methods instead of
		tossing packets over the network.)

		It's not sufficient to simply look for local hardware access,
		since it's plausible that Python on one cryo board could
		instantiate a CryoBoard object for another board. (OK, it's
		less plausible here than for a DfMUX board, but we're trying
		to keep the Python code at least similar.)
		'''

		# Macro-ize.
		if hasattr(self,'_is_local'): return self._is_local

		remoteip = self.get_ip()

		# Create a socket pointing at the board. If the outgoing
		# interface has the same IP address as the board, we're local
		# too. The port is basically arbitrary.
		port = DEFAULT_CGI_PORT
		s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
		s.connect((remoteip, port))
		localip, localport = s.getsockname()
		s.close()

		if(remoteip == localip):
			self._is_local = True
			return self._is_local

		# Out of ideas. Assume it's a remote device.
		self._is_local = False
		return self._is_local


class CryoRemoteProxy(object):
	'''
	Remote proxy for device access. Sorry, there's some more black magic
	in here.

	Basically, we assume any access to this class's functions (other than
	a handful of local variables) constitutes a remote method call. Using
	the name of the function, we form a URL. The remaining arguments are
	passed as a JSON object to the remote web server.
	'''

	def __init__(self, hostname, port):
		self._hostname = hostname
		self._port = port


	def __getattribute__(self, name):
		'''
		This function is called to get attributes (e.g. class
		variables and functions). We hijack it.
		'''

		try: return object.__getattribute__(self, name)
		except AttributeError: pass

		invoke = object.__getattribute__(self, '_invoke')
		return lambda *x, **y: invoke(name, *x, **y)


	def _invoke(self, method, *pargs, **kwargs):
		'''
		Here's where remote method calls actually take place.
		'''

		# Validation is a bit futile here, since we may run into
		# method-specific errors anyway. We try to pass exceptions
		# back and forth so that the errors are at least readable.

		encoded_args = json.dumps({
			"varargs": list(pargs),
			"kwargs": kwargs,
			"object" : "CryoBoard"
		})

		baseurl = 'http://%s:%i/%s.cgi' % (self._hostname, self._port, method)
		try: wrapped_result = json.loads(urllib.urlopen(baseurl, encoded_args).read())
		except IOError, e:
			raise CryoRemoteError("Unable to contact remote cryo board!")

		if 'error' in wrapped_result:
			raise CryoRemoteError(wrapped_result['error']['trace'])

		result = wrapped_result['result']
		return result


	def is_local(self):
		return False


	def get_converter(self):
		'''
		This function has to be overloaded here since it returns an
		object that can't be funneled through JSON.
		'''
		return CryoConverter(self)


	def get_heater_converter(self):
		'''
		This function has to be overloaded here since it returns an
		object that can't be funneled through JSON.
		'''
		return CryoHeaterConverter(self)

		
class CryoLocalProxy(object):
	'''
	Local proxy for device access.

	Note that *every* device interaction must be supported here, since
	it's how the webserver interacts with hardware. Otherwise, it would be
	tempting to simply accelerate the common calls and fall back on
	network interaction for everything else.
	'''

	# An (un-globbed) path to the cryo device structure. Note that we're
	# allowed to be sloppy about path separators since we're always
	# running this code on the cryo board, which is always Linux.
	_path = '/sys/bus/of_platform/devices/*.cryo-sequencer/'

	def __init__(self):
		path = glob.glob(self._path)
		if len(path) != 1:
			raise CryoInternalError("Can't find /sys/path_to_cryo!")
		path = path[0]
		self._path = path

		# Check the device magic
		magic_str = open(path + 'magic').read()
		if magic_str != '0x6372796f\n':
			raise CryoInternalError("Cryo magic mismatch!")

		# Embed conversion objects
		self._converter = CryoConverter(self)
		self._heater_converter = CryoHeaterConverter(self)


	def is_local(self):
		return True


	def get_heater_converter(self):
		return self._heater_converter


	def get_converter(self):
		return self._converter


	def get_converter_data(self):
		return self._converter.get_converter_data()


	def get_heater_converter_data(self):
		return self._heater_converter.get_converter_data()


	@require("frequency", tuple, list)
	def set_frequency(self, frequency):
		if len(frequency) != 2:
			raise CryoArgumentError("Frequency must be a (val, units) tuple!")

		freq_raw = self._converter.convert_frequency(
			frequency[0], frequency[1],
			Units.RAW)

		open(self._path + 'seq/freq','w').write('%i\n' % freq_raw)


	def get_frequency(self, units=Units.HZ):
		freq_raw_str = open(self._path + 'seq/freq').read()
		freq_raw = string.atoi(freq_raw_str)
		return self._converter.convert_frequency(
			freq_raw,
			Units.RAW,
			units)


	@require("channel", int, None)
	@require("amplitude", tuple, list, None)
	@require("mod_signal", SQUARE, SINE, DC, None)
	@require("demod_signal", SINE, DC, None)
	@require("adc_offset", int, float, None)
	def set_channel_parameters(self, channel,
			amplitude=None,
			mod_signal=None,
			demod_signal=None,
			adc_offset=None):

		if channel<0 or channel>=NUM_COOKED_CHANNELS:
			raise CryoArgumentError("Invalid channel %i" % channel)
		path = self._path + 'seq/channels/%i/' % channel

		if amplitude is not None:
			if len(amplitude) != 2:
				raise CryoArgumentError("Amplitude must be a (val, units) tuple!")

			amp_raw = self._converter.convert_amplitude(
					amplitude[0],
					amplitude[1],
					Units.DAC_COUNTS)

			open(path + 'amp','w').write('%i' % amp_raw)

		if mod_signal is not None:
			if mod_signal == SQUARE:
				open(path + 'dac_mixer_bypass','w').write('0')
				open(path + 'dac_mixer_square','w').write('1')
			elif mod_signal == SINE:
				open(path + 'dac_mixer_bypass','w').write('0')
				open(path + 'dac_mixer_square','w').write('0')
			elif mod_signal == DC:
				open(path + 'dac_mixer_bypass','w').write('1')
				open(path + 'dac_mixer_square','w').write('0')

		if demod_signal is not None:
			if demod_signal == SINE:
				open(path + 'adc_mixer_bypass','w').write('0')
			elif demod_signal == DC:
				open(path + 'adc_mixer_bypass','w').write('3')

		if adc_offset is not None:
			open(path + 'adc_offset','w').write('%i' % adc_offset)


	@require("channel", int)
	def get_channel_parameters(self, channel, amplitude_units=Units.DAC_COUNTS):

		if channel<0 or channel>=NUM_COOKED_CHANNELS:
			raise CryoArgumentError("Invalid channel %i" % channel)
		path = self._path + 'seq/channels/%i/' % channel

		amp = string.atoi(open(path + 'amp').read())
		dac_mixer_bypass = string.atoi(open(path + 'dac_mixer_bypass').read())
		dac_mixer_square = string.atoi(open(path + 'dac_mixer_square').read())
		adc_mixer_bypass = string.atoi(open(path + 'adc_mixer_bypass').read())
		adc_offset = string.atoi(open(path + 'adc_offset').read())

		amp = self._converter.convert_amplitude(amp, Units.DAC_COUNTS, amplitude_units)

		result = {
			'amplitude' : (amp, amplitude_units),
			'adc_offset' : adc_offset
		}

		if dac_mixer_bypass==1: result['mod_signal'] = DC
		elif dac_mixer_square==1: result['mod_signal'] = SQUARE
		else: result['mod_signal'] = SINE

		if adc_mixer_bypass==1: result['demod_signal'] = DC
		else: result['demod_signal'] = SINE

		return result


	@require("pga", int, PGA_GAIN_TO_REG_MAP.keys())
	@require("buffer", bool)
	def set_adc_parameters(self, pga=None, buffer=None):
		'''
		Set ADC parameters.

		Note that these parameters affect device behaviour in sneaky
		ways. For example, the PGA changes the "zero" voltage setting,
		which requires you to modify each channel's offset register to
		propery measure DC levels.
		'''

		# TODO: race conditions? ADC SPI writes must not be too
		# closely spaced.

		# TODO: autocalibration?

		file = self._path + 'seq/adc_spi_ctl'

		if pga is not None:
			reg = PGA_GAIN_TO_REG_MAP[pga]
			open(file,'w').write('0x0f0002%02x' % reg)

		if buffer is not None:
			reg = 2 if bool else 0
			open(file,'w').write('0x0f0000%02x' % reg)


	def __streamer_command(self, msg):
		'''
		Send a command and get a response from the streamer's control
		interface. Your code should never call this directly; see the
		streamer methods for an example of the kind of thing you
		should be using.
		'''
		s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		s.connect("/tmp/decimator_ctl")

		# Receive greeting, validate it
		greeting = s.recv(64)
		match = re.match("HELO CRYO STREAMER V([0-9]*)", greeting)
		if match is None:
			raise CryoRemoteError("Can't parse streamer version!")
		version = int(match.expand(r"\1"))

		# Send launch command; read response
		s.send(msg)
		response = s.recv(1024).splitlines()
		s.close()

		return response


	@require("ip", str,unicode)
	@require("port", int)
	def launch_streamer(self, ip, port):
		'''
		Launch streamer
		'''
		r = self.__streamer_command("stream %s %i" % (ip,port))
		if r[-1] != "OK":
			raise CryoArgumentError("Can't stream to requested address/port (%s)" % r[-1])


	@require("ip", str,unicode)
	@require("port", int)
	def kill_streamer(self, ip, port):
		'''
		Kill streamer
		'''
		r = self.__streamer_command("kill %s %i" % (ip,port))
		if r[-1] != "OK":
			raise CryoArgumentError("Can't kill requested streamer")


	def get_streamers(self):
		'''
		Find out which streamers are running
		'''
		r = self.__streamer_command("list")
		if r[-1] != "OK":
			raise CryoInternalError("Error asking for streamer list!")
		del r[-1]

		streamers = []

		for line in r:
			line = line.split(' ')
			if line[0] == 'INET':
				streamers.append({
					'ip': line[1],
					'port': int(line[2])
				})

		return streamers


	@require("channel", int)
	@require("amplitude", int, float, tuple, list, None)
	@require("switch", bool, None)
	def set_heater_parameters(self, channel,
			amplitude=None,
			switch=None):

		if not isinstance(amplitude, (tuple, list)) or len(amplitude) != 2:
			raise CryoArgumentError("Expecting (value, units) tuple for amplitude argument.")
		if channel<0 or channel>=NUM_HEATER_CHANNELS:
			raise CryoArgumentError("Invalid heater channel %i" % channel)
		path = self._path + 'heater/channels/%i/' % channel

		if amplitude is not None:
			amp_raw = self._heater_converter.convert_amplitude(
					amplitude[0], amplitude[1],
					Units.DAC_COUNTS)

			open(path + 'amp','w').write('%i' % amp_raw)

		if switch is not None:
			if switch: switch=1
			else: switch=0

			open(path + 'switch','w').write('%i' % switch)


	@require("channel", int)
	@require("amplitude_units", Units.DAC_COUNTS, Units.VOLTS)
	def get_heater_parameters(self, channel, amplitude_units = Units.DAC_COUNTS):

		if channel<0 or channel>=NUM_HEATER_CHANNELS:
			raise CryoArgumentError("Invalid heater channel %i" % channel)
		path = self._path + 'heater/channels/%i/' % channel

		amp = string.atoi(open(path + 'amp').read())
		switch = string.atoi(open(path + 'switch').read())

		amp = self._heater_converter.convert_amplitude(amp, Units.DAC_COUNTS, amplitude_units)

		if switch==1: switch=True
		else: switch=False

		result = {
			'amplitude' : (amp, amplitude_units),
			'switch' : switch
		}

		return result


	def get_rtl_version(self):
		'''
		Returns the version register supplied by the RTL
		'''
		return string.atoi(open(self._path + 'version').read())


	def get_crython_version(self):
		'''
		Returns the version number supplied by the Crython class
		'''
		return CRYTHON_VERSION


	def get_crython_commit(self):
		'''
		Returns the Git HEAD commit for the crython repository
		'''
		return open('/home/crython/.git/refs/heads/master').read()


	def reset_board(self):
		'''
		Resets board parameters to some consistent (and hopefully
		safe) state.
		'''

		self.set_frequency(frequency=0)

		for c in xrange(NUM_COOKED_CHANNELS):
			self.set_channel_parameters(
				channel=c,
				amplitude=32767,
				mod_signal=SINE,
				demod_signal=SINE,
				adc_offset=-2097152)

		for c in xrange(NUM_HEATER_CHANNELS):
			self.set_heater_parameters(
				channel=c,
				amplitude=0,
				switch=False)


class CryoStreamer(object):

	def __init__(self, cryoboard, ip, port):

		# Validate arguments and store them. We permit cryoboard to be
		# None (if we have a streamer URL and don't want to supply the
		# board URL), but launch/kill/status methods will be
		# unavailable.
		if cryoboard is not None and not isinstance(cryoboard, CryoBoard):
			raise CryoArgumentError("Invalid cryo board")

		if not isinstance(ip, str):
			raise CryoArgumentError("Invalid IP")

		if not isinstance(port, int):
			raise CryoArgumentError("Invalid port")

		self._board = cryoboard
		(self._ip, self._port) = (ip, port)


	def launch(self):
		if self._board is None: raise CryoArgumentError('''
			Can't launch streamer without a CryoBoard!  (hint: you
			aren't instantiating this CryoStreamer correctly)
		''')
		self._board.launch_streamer(self._ip, self._port)


	def kill(self):
		if self._board is None: raise CryoArgumentError('''
			Can't launch streamer without a CryoBoard!  (hint: you
			aren't instantiating this CryoStreamer correctly)
		''')
		self._board.kill_streamer(self._ip, self._port)


	def status(self):
		'''
		Returns True if the streamer is running; False otherwise
		'''

		if self._board is None: raise CryoArgumentError('''
			Can't launch streamer without a CryoBoard!  (hint: you
			aren't instantiating this CryoStreamer correctly)
		''')

		for s in self._board.get_streamers():
			if s['ip'] == self._ip and s['port'] == self._port:
				return True

		return False


	def get_samples(self, num_samples):
		try:	s = self.__socket
		except AttributeError:
			s = self.__socket = socket.socket(
				socket.AF_INET,
				socket.SOCK_DGRAM, socket.IPPROTO_UDP
			)
			s.bind((self._ip, self._port))

		# Parse as a struct; convert into a list
		raw_samples = []

		x = struct.Struct('!II64i')
		for samp in xrange(num_samples):
			data,addr = s.recvfrom(x.size)
			y = x.unpack(data)
			if y[0] != 0x4352594f:
				print y[0]
				raise CryoInternalError("Got bad packet from streamer!")
			if y[1] < 1:
				raise CryoInternalError("Got packet from unsupported streamer version!")
			raw_samples.append(y[2:])

		# Invert the list: we wanted raw_samples[chan][sample], but
		# got raw_samples[sample][chan]
		raw_samples = [ [ raw_samples[s][c] for s in xrange(num_samples) ] for c in xrange(NUM_RAW_CHANNELS) ]

		# Convert into dictionary of floats
		out = {
			"drive": {
				"i": raw_samples[0:NUM_RAW_CHANNELS:4],
				"q": raw_samples[1:NUM_RAW_CHANNELS:4],
			},
			"sense": {
				"i": raw_samples[2:NUM_RAW_CHANNELS:4],
				"q": raw_samples[3:NUM_RAW_CHANNELS:4],
			}
		}

		# Add RMS conversion
		rmsf = lambda x,y: math.sqrt(x**2 + y**2)
		for who, what in out.iteritems():
			what["rms"] = [ map(rmsf,x,y) for (x,y) in zip(what['i'], what['q']) ]
			
		return out


'''
Conversion classes follow. The pattern is as follows:

   from_units -> raw (machine) units -> to_units

We use machine units (e.g. DAC/ADC counts) as the "currency" here since it's
what firmware uses.  As a side effect, this function can be used to figure out
what you "really got" when you asked for some number in "human units."
'''

class CryoHeaterConverter(object):
	'''
	Heater conversions.
	'''

	CONFIG_FILE = '/etc/cryo/heater_config'

	def __init__(self, cryo_board):
		self._cb = cryo_board
		self.load()


	def load(self):
		if self._cb.is_local():
			config = ConfigParser.ConfigParser()
			config.read(self.CONFIG_FILE)

			self.config = {
				'supply_voltage': config.getfloat('Global', 'supply_voltage'),
				'Names': [ config.get('Names', str(c))
					for c in xrange(NUM_HEATER_CHANNELS) ],
				'Resistances': [ config.getfloat('Resistances', str(c))
					for c in xrange(NUM_HEATER_CHANNELS) ]
			}
		else:
			self.config = self._cb.get_converter_data()


	def save(self):
		if not self._cb.is_local():
			raise CryoInternalError("Not supported (yet)!")

		config = ConfigParser.ConfigParser()

		for section in ('Global', 'Names', 'Resistances'):
			config.add_section(section)

		config.set('Global', 'supply_voltage', str(self.config['supply_voltage']))

		for c in xrange(NUM_HEATER_CHANNELS):
			config.set('Names', str(c), self.config['Names'][c])
			config.set('Resistances', str(c), str(self.config['Resistances'][c]))

		f = open(self.CONFIG_FILE, 'w')
		config.write(f)
		f.close()


	def get_converter_data(self):
		return self.config


	@require("channel", *range(NUM_HEATER_CHANNELS))
	@require("name", None, str, unicode)
	@require("resistance", None, int, float)
	def set_channel_parameters(self, channel,
			name=None, resistance=None):

		if name: self.config['Names'][channel] = name
		if resistance: self.config['Resistances'][channel] = resistance

		self.save()


	@require("amp", int, float, list, tuple)
	@require("from_units", Units.DAC_COUNTS, Units.VOLTS)
	@require("to_units", Units.DAC_COUNTS, Units.VOLTS)
	def convert_amplitude(self, amp, from_units, to_units, channel=None):
	
		unwrap = isinstance(amp, (float, int))
		if unwrap: amp = [amp]

		# Convert input units to raw units
		if from_units == Units.DAC_COUNTS:
			pass
		elif from_units == Units.VOLTS:
			amp = map(lambda x: int(round( (x*65535.) / self.config['supply_voltage'] )), amp)

		# Remove input units; validate raw units
		if any(map(lambda x: x < 0 or x > 65535, amp)):
			raise CryoArgumentError("Machine units out of range!")

		if to_units == Units.DAC_COUNTS:
			pass
		elif to_units == Units.VOLTS:
			amp = map(lambda x: (x/65535.) * self.config['supply_voltage'], amp)

		return amp.pop() if unwrap else amp


class CryoConverter(object):

	CONFIG_FILE='/etc/cryo/therm_config'

	def __init__(self, cryo_board):
		self._cb = cryo_board
		self.load()


	def load(self):
		if self._cb.is_local():
			config = ConfigParser.ConfigParser()
			config.read(self.CONFIG_FILE)

			r = xrange(NUM_COOKED_CHANNELS)

			# Parse config entries
			self.config = {
				'Names': [ config.get('Names', str(c)) for c in r ],
				'Types': [ config.get('Types', str(c)) for c in r ],
				'Resistances': [ config.getfloat('Resistances', str(c)) for c in r ],
				'Calibrations': [ config.get('Calibrations', str(c)) for c in r ]
			}

			# Validate config entries
			if not all(map(lambda x: x in ["RTD", "Diode"], self.config['Types'])):
				raise CryoCalibrationError("Unknown sensor type in calibration file!")

			if not all(map(os.path.isfile, self.config['Calibrations'])):
				raise CryoCalibrationError("Calibration file didn't exist!")

			# Parse calibrations
			res2temp_t = []
			res2temp_r = []
			for c in r:
				f = open(self.config['Calibrations'][c])
				lines = f.readlines()
				f.close()

				(t,r) = zip(*[ map(float, l.split()) for l in lines ])
				res2temp_t.append(t)
				res2temp_r.append(r)

			self.config['res2temp_t'] = res2temp_t
			self.config['res2temp_r'] = res2temp_r

		else:
			self.config = self._cb.get_converter_data()


	def save(self):
		if not self._cb.is_local():
			raise CryoInternalError("Not supported (yet)!")

		config = ConfigParser.ConfigParser()

		for section in ('Names', 'Types', 'Resistances', 'Calibrations'):
			config.add_section(section)

		for c in xrange(NUM_COOKED_CHANNELS):
			config.set('Names', str(c), self.config['Names'][c])
			config.set('Types', str(c), self.config['Types'][c])
			config.set('Resistances', str(c), str(self.config['Resistances'][c]))
			config.set('Calibrations', str(c), str(self.config['Calibrations'][c]))

		f = open(self.CONFIG_FILE, 'w')
		config.write(f)
		f.close()


	@require("channel", int)
	@require("name", str, unicode)
	@require("type", None, "RTD", "Diode")
	@require("resistance", None, int, float)
	@require("calibration", None) # FIXME
	def set_channel_parameters(self, channel,
			name=None, type=None, resistance=None, calibration=None):

		if name: self.config['Names'][channel] = name
		if type: self.config['Types'][channel] = type
		if resistance: self.config['Resistances'][channel] = resistance

		if calibration:
			#f = fopen(self.config['Calibrations'][channel], 'w')
			#f.write(calibration)
			#f.close()
			raise CryoInternalError("Not supported (yet!)")

		self.save()


	def get_converter_data(self):
		return self.config


	@require("freq", int, float, list, tuple)
	@require("from_units", Units.HZ, Units.RAW)
	@require("to_units", Units.HZ, Units.RAW)
	def convert_frequency(self, freq, from_units, to_units):

		unwrap = isinstance(freq, (float, int))
		if unwrap: freq = [ freq ]

		if from_units == Units.HZ:
			freq = map(lambda x: round((x / FSAMP) * 2**32), freq)
		elif from_units == Units.RAW:
			pass

		if any(map(lambda x: x < 0 or x >= 2**31, freq)):
			raise CryoArgumentError("Negative frequency or frequency above Nyquist!")

		if to_units == Units.HZ:
			freq = map(lambda x: (x / 2.**32) * FSAMP, freq)
		elif to_units == Units.RAW:
			pass

		return freq.pop() if unwrap else freq


	@require("amplitude", int, float, list, tuple)
	@require("from_units", Units.DAC_COUNTS, Units.VOLTS)
	@require("to_units", Units.DAC_COUNTS, Units.VOLTS)
	def convert_amplitude(self, amplitude, from_units, to_units):

		unwrap = isinstance(amplitude, (float, int))
		if unwrap: amplitude = [ amplitude ]

		# Convert input units to raw units
		if from_units == Units.DAC_COUNTS:
			pass
		elif from_units == Units.VOLTS:
			amplitude = map(lambda x: int(round(x/self._dac2volt_cal[1] + self._dac2volt_cal[0])), amplitude)

		# Validate raw units
		if any(map(lambda x: x < 0 or x > 65535, amplitude)):
			raise CryoArgumentError("Machine units out of range!")

		if to_units == Units.DAC_COUNTS:
			pass
		elif to_units == Units.VOLTS:
			amplitude = map(lambda x: (x-self._dac2volt_cal[0])*self._dac2volt_cal[1], amplitude)
			amplitude = [ (a if a >= 0 else 0) for a in amplitude ]

		return amplitude.pop() if unwrap else amplitude


	@require("temp", int, float, list, tuple)
	@require("from_units", Units.ADC_COUNTS, Units.VOLTS, Units.OHMS, Units.KELVIN)
	@require("to_units", Units.ADC_COUNTS, Units.VOLTS, Units.OHMS, Units.KELVIN)
	def convert_temperature(self, temp, channel, from_units, to_units):

		# Conversion matrix. This is indexed by the "source" units,
		# furnishing a dictionary indexed by "destination" units. It
		# doesn't always give you the units you asked for, but it must
		# proceed in the correct direction.
		_cm = {
			Units.ADC_COUNTS: {
				Units.VOLTS:		( self._adc2volt, Units.VOLTS),
				Units.OHMS:		( self._adc2volt, Units.VOLTS),
				Units.KELVIN:		( self._adc2volt, Units.VOLTS)
			},
			Units.VOLTS: {
				Units.ADC_COUNTS:	( self._volt2adc, Units.ADC_COUNTS),
				Units.OHMS:		( self._volt2ohm, Units.OHMS),
				Units.KELVIN:		( self._volt2ohm, Units.OHMS)
			},
			Units.OHMS: {
				Units.ADC_COUNTS:	( self._ohm2volt, Units.VOLTS),
				Units.VOLTS:		( self._ohm2volt, Units.VOLTS),
				Units.KELVIN:		( self._ohm2temp, Units.KELVIN)
			},
			Units.KELVIN: {
				Units.ADC_COUNTS:	( self._temp2ohm, Units.OHMS),
				Units.VOLTS:		( self._temp2ohm, Units.OHMS),
				Units.OHMS:		( self._temp2ohm, Units.OHMS)
			}
		}

		unwrap = isinstance(temp, (float, int))
		if unwrap: temp = [ temp ]

		while from_units != to_units:
			temp = map( lambda t: _cm[from_units][to_units][0](channel,t), temp)
			from_units = _cm[from_units][to_units][1]

		return temp.pop() if unwrap else temp


	# Calibrations are (offset, slope) tuples
	_adc2volt_cal = (0, 10./pow(2.,24.))
	_dac2volt_cal = (-pow(2.,15.), 2.5/pow(2.,16.))
	_dac_bias = 0

	# Helper functions... don't call these from user code

	def _get_bias_current(self, channel):
		r = self.config['Resistances'][channel]
		vbias = self.convert_amplitude(self._dac_bias, Units.DAC_COUNTS, Units.VOLTS)
		return vbias / r

	def _adc2volt(self,channel,adc):
		volt=(adc-self._adc2volt_cal[0]) * self._adc2volt_cal[1]
		return volt

	def _volt2adc(self,channel,volt):
		adc=volt/self._adc2volt_cal[1] + self._adc2volt_cal[0]
		return adc

	def _volt2ohm(self,channel,volt):
		res=volt/self._get_bias_current(channel)
		return res

	def _ohm2volt(self,channel,res):
		volt=res*self._get_bias_current(channel)
		return volt

	def _temp2ohm(self,channel,temp):
		try:
			res=self._interp1d(self.config['res2temp_t'][channel],self.config['res2temp_r'][channel],temp)
			return res
		except CryoCalibrationError:
			return 1e20

	def _ohm2temp(self,channel,res):
		try:
			temp=self._interp1d(self.config['res2temp_r'][channel],self.config['res2temp_t'][channel],res)
			return res
		except CryoCalibrationError:
			return 1e20


	def _interp1d(self, xs,ys,x):

		for (x1,x2), (y1,y2) in zip(zip(xs[:-1],xs[1:]), zip(ys[:-1],ys[1:])):
			if x >= min(x1,x2) and x <= max(x1,x2):
				return ( y1*(x2-x) + y2*(x-x1) ) / (x2-x1)

		raise CryoCalibrationError("Out of range!")

